// --------------------------------------------------------------------
// Canvas
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2003  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "ipepage.h"
#include "ipestyle.h"

#include "qipe.h"
#include "canvas.h"
#include "overlay.h"

#include <qpainter.h>
#include <qstatusbar.h>

// --------------------------------------------------------------------

/*! \class Canvas
  \brief A widget that displays an Ipe document page.
*/

//! Construct a new canvas.
Canvas::Canvas(OverlayProvider *provider, QStatusBar *statusBar,
	       QWidget* parent, WFlags f)
  : QWidget(parent, "canvas", f), iStatusBar(statusBar), iProvider(provider)
{
  assert(iProvider);
  setBackgroundMode(NoBackground);
  iPan = IpeVector::Zero;
  iZoom = 1.0;
  iDimmed = false;
  iDoubleBuffer = false;
  iOffscreen.resize(size());
  iRepaintObjects = false;
  iOverlay = 0;
  setFocusPolicy(ClickFocus);
  setMouseTracking(true);
  iAutoSnap = false;
  iFifiVisible = false;
}

//! Canvas destructor.
Canvas::~Canvas()
{
  delete iOverlay;
}

// --------------------------------------------------------------------

//! Set the page to be displayed.
/*! Doesn't take ownership of any argument. */
void Canvas::SetPage(const IpePage *page, const IpeStyleSheet *sheet,
		     const PdfFontPool *fontPool)
{
  iPage = page;
  iStyleSheet = sheet;
  iFontPool = fontPool;
  delete iOverlay;
  iOverlay = 0;
  Update();
}

//! Set current pan position.
/*! The pan position is the user coordinate that is displayed at
  the very center of the canvas. */
void Canvas::SetPan(const IpeVector &v)
{
  iPan = v;
  Update();
}

//! Set current zoom factor.
/*! The zoom factor maps user coordinates to screen pixel coordinates. */
void Canvas::SetZoom(IpeScalar zoom)
{
  iZoom = zoom;
  Update();
}

//! Set the snapping information.
void Canvas::SetSnap(const IpeSnapData &s)
{
  iSnapData = s;
}

//! Dim whole canvas, except for the overlay.
/*! This mode will be reset when the overlay finishes. */
void Canvas::SetDimmed(bool dimmed)
{
  iDimmed = dimmed;
  Update();
}

//! Set double buffering for overlay.
/*! This mode will be reset when the overlay finishes. */
void Canvas::SetDoubleBuffer(bool db)
{
  iDoubleBuffer = db;
}

//! Enable automatic angular snapping with this origin.
/*! Snapping is disabled automatically on FinishOverlay(). */
void Canvas::SetAutoOrigin(const IpeVector &v)
{
  iAutoOrigin = v;
  iAutoSnap = true;
}

//! Mark for update with redrawing of objects.
void Canvas::Update()
{
  iRepaintObjects = true;
  update();
}

//! Mark for update with redrawing of overlay only.
void Canvas::UpdateOverlay()
{
  update();
}

//! Clear status bar.
void Canvas::ClearMessage()
{
  iStatusBar->clear();
}

//! Display a message in the status bar.
void Canvas::Message(QString str)
{
  iStatusBar->message(str);
}

IpeVector Canvas::DevToUser(const IpeVector &arg) const
{
  IpeVector v = arg - Center();
  v.iX /= iZoom;
  v.iY /= -iZoom;
  v += iPan;
  return v;
}

IpeVector Canvas::UserToDev(const IpeVector &arg) const
{
  IpeVector v = arg - iPan;
  v.iX *= iZoom;
  v.iY *= -iZoom;
  v += Center();
  return v;
}

//! Adjust painter for canvas coordinate system.
IpeMatrix Canvas::CanvasTfm() const
{
  return IpeMatrix(Center()) *
    IpeLinear(iZoom, 0, 0, -iZoom) *
    IpeMatrix(-iPan);
}

// --------------------------------------------------------------------

void Canvas::SetOverlay(Overlay *overlay)
{
  iOverlay = overlay;
  iProvider->SetDrawingMode(true);
  UpdateOverlay();
}

//! Current overlay has done its job.
/*! Overlay is deleted, canvas fully updated, and cursor reset. */
void Canvas::FinishOverlay()
{
  delete iOverlay;
  iOverlay = 0;
  iDimmed = false;
  iAutoSnap = false;
  iDoubleBuffer = false;
  Update();
  unsetCursor();
  iProvider->SetDrawingMode(false);
  ClearMessage();
}

/*! Stores the mouse position in iUnsnappedMousePos, computes Fifi if
  snapping is enabled, and stores snapped position in iMousePos. */
void Canvas::ComputeFifi(QMouseEvent *ev)
{
  iUnsnappedMousePos = DevToUser(IpeVector(ev->x(), ev->y()));
  iMousePos = iUnsnappedMousePos;

  int mask = iAutoSnap ? 0 : IpeSnapData::ESnapAuto;
  if (iSnapData.iSnap & ~mask) {
    iSnapData.Snap(iMousePos, iPage, iSnapData.iDistance / iZoom,
		   (iAutoSnap ? &iAutoOrigin : 0));

    // convert fifi coordinates back into device space
    QPoint qfifi = QPt(UserToDev(iMousePos));
    if (iFifi != qfifi) {
      // update old and new fifi position
      update(iFifi.x() - 8, iFifi.y() - 8, 17, 17);
      iFifi = qfifi;
      update(iFifi.x() - 8, iFifi.y() - 8, 17, 17);
    }
  } else if (iFifiVisible) {
    iFifiVisible = false;
    update(iFifi.x() - 8, iFifi.y() - 8, 17, 17);
  }
}

// --------------------------------------------------------------------

void Canvas::DrawAxes(QPainter &qPainter)
{
  IpeMatrix m = CanvasTfm() * IpeMatrix(iSnapData.iOrigin);

  qPainter.setPen(Qt::green);
  int x = int(m.Translation().iX);
  int y = int(m.Translation().iY);

  double alpha = 0.0;
  IpeVector ep((width() + height()) / iZoom, 0);

  while (alpha < IpeTwoPi) {
    double beta = iSnapData.iDir + alpha;
    qPainter.moveTo(x, y);
    qPainter.lineTo(QPt(m * IpeLinear(beta) * ep));
    alpha += iSnapData.iAngleSize;
  }

  if (iSnapData.iAngleSize > (0.9 * IpePi))
    qPainter.drawEllipse(x - 2, y - 2, 4, 4);
}

void Canvas::DrawGrid(QPainter &qPainter, const IpeRect &media)
{
  double step = iSnapData.iGridSize;
  if (step * iZoom < 2.0)
    return;

  int vfactor = int(1.0 + 16.0/(step * iZoom));
  double vstep = step * vfactor;

  IpeVector ll = media.Min();
  IpeVector ur = media.Max();

  ll.iX = vstep * int(ll.iX / vstep);
  ll.iY = vstep * int(ll.iY / vstep);

  IpeMatrix m = CanvasTfm();

  IpeVector v;
  qPainter.setPen(Qt::black);
  for (v.iX = ll.iX; v.iX < ur.iX; v.iX += vstep)
    for (v.iY = ll.iY; v.iY < ur.iY; v.iY += step)
      qPainter.drawPoint(QPt(m * v));
  if (vfactor != 1) {
    for (v.iY = ll.iY; v.iY < ur.iY; v.iY += vstep)
      for (v.iX = ll.iX; v.iX < ur.iX; v.iX += step)
	qPainter.drawPoint(QPt(m * v));
  }
}

void Canvas::RepaintObjects()
{
  iOffscreen.fill(Qt::lightGray);
  iRepaintObjects = false;
  if (!iPage)
    return;

  QPainter qPainter;
  qPainter.begin(&iOffscreen);

  CanvasPainter painter(iStyleSheet, iFontPool, &iOffscreen,
			&qPainter, iZoom, false);
  painter.Transform(CanvasTfm());
  painter.Push();
  painter.SetStroke(IpeAttribute::Void());
  painter.SetFill(iProvider->PageColor());
  painter.Rect(iPage->MediaBox());
  painter.DrawPath();
  painter.Pop();
  painter.SetDimmed(iDimmed);

  if (iSnapData.iGridVisible)
    DrawGrid(qPainter, iPage->MediaBox());
  if (iSnapData.iWithAxes)
    DrawAxes(qPainter);

  painter.Push();
  for (IpePage::const_iterator it = iPage->begin(); it != iPage->end(); ++it) {
    const IpeLayer &l = iPage->Layer(it->Layer());
    if (l.IsVisible()) {
      painter.SetDimmed(iDimmed || l.IsDimmed());
      it->Object()->Draw(painter);
    }
  }
  painter.Pop();
  qPainter.end();
}

// --------------------------------------------------------------------

void Canvas::paintEvent(QPaintEvent *ev)
{
  if (iRepaintObjects)
    RepaintObjects();
  //QRect rect = ev->rect();

  QPixmap buffer;
  QPainter qPainter;

  if (iDoubleBuffer) {
    buffer.resize(size());
    bitBlt(&buffer, 0, 0, &iOffscreen,
	   0, 0, iOffscreen.width(), iOffscreen.height());
    qPainter.begin(&buffer);
  } else {
    bitBlt(this, 0, 0, &iOffscreen,
	   0, 0, iOffscreen.width(), iOffscreen.height());
    qPainter.begin(this);
  }

  if (iOverlay) {
    iOverlay->Draw(ev, &qPainter);
  } else {
    OverlayPainter painter(iStyleSheet, &qPainter);
    painter.Transform(CanvasTfm());

    QPen primaryPen;
    QPen secondaryPen;
    primaryPen.setColor(Qt::red);
    primaryPen.setWidth(2);
    secondaryPen.setColor(QColor("purple"));

    painter.Push();
    for (IpePage::const_iterator it = iPage->begin();
	 it != iPage->end(); ++it) {
      if (it->Select() == IpePgObject::EPrimary) {
	qPainter.setPen(primaryPen);
	it->Object()->Draw(painter);
      } else if (it->Select() == IpePgObject::ESecondary) {
	qPainter.setPen(secondaryPen);
	it->Object()->Draw(painter);
      }
    }
    painter.Pop();
  }

  if (iSnapData.iSnap) {
    qPainter.setPen(Qt::red);
    qPainter.drawLine(iFifi.x() - 8, iFifi.y(), iFifi.x() + 8, iFifi.y());
    qPainter.drawLine(iFifi.x(), iFifi.y() - 8, iFifi.x(), iFifi.y() + 8);
    iFifiVisible = true;
  }
  qPainter.end();
  if (iDoubleBuffer)
    bitBlt(this, 0, 0, &buffer, 0, 0, buffer.width(), buffer.height());
}

void Canvas::resizeEvent(QResizeEvent *ev)
{
  iOffscreen.resize(size());
  iRepaintObjects = true;
}

void Canvas::wheelEvent(QWheelEvent *ev)
{
  iProvider->WheelZoom(ev->delta());
  ev->accept();
}

void Canvas::mousePressEvent(QMouseEvent *ev)
{
  ClearMessage();
  ComputeFifi(ev);
  if (iOverlay) {
    iOverlay->MousePress(ev);
  } else {
    // we need to save overlay in a temporary variable, since
    // it is possible that function returns 0, but we already
    // have another overlay (user discards object menu by clicking
    // on canvas).
    iProvider->RequestOverlay(ev);
  }
}

void Canvas::mouseReleaseEvent(QMouseEvent *ev)
{
  ComputeFifi(ev);
  if (iOverlay)
    iOverlay->MouseRelease(ev);
}

void Canvas::mouseMoveEvent(QMouseEvent *ev)
{
  ComputeFifi(ev);
  if (iOverlay)
    iOverlay->MouseMove(ev);
}

void Canvas::keyPressEvent(QKeyEvent *ev)
{
  ClearMessage();
  // Escape aborts overlay
  if (ev->key() == Key_Escape && iOverlay)
    FinishOverlay();
  else if (iOverlay)
    iOverlay->KeyPress(ev);
  else
    ev->ignore();
}

QSize Canvas::sizeHint() const
{
  return QSize(640, 480);
}

// --------------------------------------------------------------------
