// -*- C++ -*-
// --------------------------------------------------------------------
// Canvas
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2003  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef CANVAS_H
#define CANVAS_H

#include "ipegeo.h"
#include "ipecolor.h"
#include "ipesnap.h"
#include "ipepainter.h"

#include <qwidget.h>
#include <qpixmap.h>
#include <qpainter.h>

// --------------------------------------------------------------------

class IpeStyleSheet;
class IpePage;
class PdfFontPool;
class Overlay;
class QMouseEvent;
class QStatusBar;

// --------------------------------------------------------------------

class CanvasPainter : public IpePainter {
public:
  CanvasPainter(const IpeStyleSheet *sheet, const PdfFontPool *fontPool,
		QPixmap *pixmap, QPainter *painter, double zoom, bool pretty);
  virtual ~CanvasPainter() { }

  void SetDimmed(bool dim) { iDimmed = dim; }

  virtual void BeginPath(const IpeVector &v);
  virtual void BeginClosedPath(const IpeVector &v);
  virtual void LineTo(const IpeVector &v);
  virtual void CurveTo(const IpeVector &v1, const IpeVector &v2,
		       const IpeVector &v3);
  virtual void EndPath();
  virtual void EndClosedPath();

  virtual void DrawPath();
  virtual void DrawBitmap(IpeBitmap bitmap);
  virtual void DrawText(const IpeText *text);

private:
  const PdfFontPool *iFontPool;

  QPixmap *iPixmap;
  QPainter *iPainter;
  double iZoom;
  bool iPretty;

  bool iDimmed;

  enum TVertexType { EBeginPath, EBeginClosedPath,
		     ELineTo, ECurveTo,
		     EEndPath, EEndClosedPath };

  std::vector<QPoint> iV;
  std::vector<TVertexType> iType;

  IpeAttribute iDash[4]; // dash, dot, dash dot, dash dot dot
};

// --------------------------------------------------------------------

/*! \class OverlayProvider
  \brief Interface for providing Overlay objects.
*/

class OverlayProvider {
public:
  //! Request an Overlay object because of this mouse press.
  virtual void RequestOverlay(QMouseEvent *ev) = 0;
  virtual void SetDrawingMode(bool mode) = 0;
  virtual void WheelZoom(int delta) = 0;
  virtual IpeAttribute PageColor() = 0;
};

class Canvas : public QWidget {
  Q_OBJECT

public:
  Canvas(OverlayProvider *provider, QStatusBar *statusBar,
	 QWidget* parent, WFlags f=0);
  ~Canvas();

  void SetPage(const IpePage *page, const IpeStyleSheet *sheet,
	       const PdfFontPool *fontPool);
  void SetPan(const IpeVector &v);
  void SetZoom(double zoom);
  void SetSnap(const IpeSnapData &s);
  void SetDimmed(bool dimmed);
  void SetAutoOrigin(const IpeVector &v);
  void SetDoubleBuffer(bool db);
  IpeMatrix CanvasTfm() const;

  //! Return current pan.
  inline IpeVector Pan() const { return iPan; }
  //! Return current zoom.
  inline double Zoom() const { return iZoom; }
  //! Return current style sheet.
  inline const IpeStyleSheet *StyleSheet() const { return iStyleSheet; }

  //! Return center of canvas.
  inline IpeVector Center() const {
    return 0.5 * IpeVector(width(), height()); }

  //! Return last mouse position (snapped!) in user coordinates.
  inline IpeVector Pos() const { return iMousePos; }

  //! Return last unsnapped mouse position in user coordinates.
  inline IpeVector UnsnappedPos() const { return iUnsnappedMousePos; }

  //! Return current snapping information.
  inline const IpeSnapData &SnapData() const { return iSnapData; }

  void Message(QString str);
  void ClearMessage();
  void Update();
  void UpdateOverlay();

  void SetOverlay(Overlay *overlay);
  void FinishOverlay();

  IpeVector DevToUser(const IpeVector &arg) const;
  IpeVector UserToDev(const IpeVector &arg) const;

protected:
  virtual void resizeEvent(QResizeEvent *ev);
  virtual void paintEvent(QPaintEvent *ev);
  virtual void mousePressEvent(QMouseEvent *ev) ;
  virtual void mouseReleaseEvent(QMouseEvent *ev);
  virtual void mouseMoveEvent(QMouseEvent *ev);
  virtual void wheelEvent(QWheelEvent *ev);
  virtual void keyPressEvent(QKeyEvent *ev);
  virtual QSize sizeHint() const;

private:
  void DrawGrid(QPainter &qPainter, const IpeRect &media);
  void RepaintObjects();
  void ComputeFifi(QMouseEvent *ev);
  void DrawAxes(QPainter &qPainter);
  void SetOrigin();
  bool SimpleSnap();

private:
  const IpePage *iPage;
  const IpeStyleSheet *iStyleSheet;
  const PdfFontPool *iFontPool;
  QStatusBar *iStatusBar;
  IpeAttribute iPageColor;

  IpeVector iPan;
  double iZoom;
  IpeSnapData iSnapData;
  bool iDimmed;
  bool iDoubleBuffer;
  bool iAutoSnap;
  IpeVector iAutoOrigin;

  OverlayProvider *iProvider;
  Overlay *iOverlay;

  bool iRepaintObjects;
  QPixmap iOffscreen;
  IpeVector iUnsnappedMousePos;
  IpeVector iMousePos;
  QPoint iFifi;
  bool iFifiVisible;
};

// --------------------------------------------------------------------
#endif
