// --------------------------------------------------------------------
// Overlay for creating IpeText objects
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2003  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "createtext.h"

#include "canvas.h"
#include "qipe.h"
#include "prefs.h"

#include <qevent.h>
#include <qpainter.h>
#include <qstring.h>
#include <qmultilineedit.h>
#include <qlabel.h>

#include "debugwin.h"

// --------------------------------------------------------------------

/*! \class DialogCreateText
  \brief Dialog to edit a text paragraph
*/

//! The argument \c data is only updated if the dialog is accepted.
DialogCreateText::DialogCreateText(QWidget* parent, QString caption,
				   QString label, QString &data)
  : DlgCreateText(parent, "dlgcreatetext", true), iData(data)
{
  setCaption(caption);
  iLabel->setText(label);
  iEditor->setText(iData);
  iEditor->setFocus();
  iEditor->setFont(SPreferences::Static()->iTextFont);
}

void DialogCreateText::UpdateData()
{
  iData = iEditor->text();
}

// --------------------------------------------------------------------

/*! \class CreateText
  \brief Overlay for creating IpeText objects.
*/

//! Constructor starts selection.
CreateText::CreateText(QMouseEvent *ev, Canvas *canvas,
		       ObjectAdder *objectAdder)
  : Overlay(canvas), iObjectAdder(objectAdder)
{
  iV[0] = iCanvas->Pos();
  iV[1] = iV[0];
}

void CreateText::New(QMouseEvent *ev, Canvas *canvas,
		     ObjectAdder *objectAdder, TMode mode)
{
  IpeVector pos = canvas->Pos();
  QString text;
  DialogCreateText *dialog =
    new DialogCreateText(canvas, QObject::tr("Create text object"),
			 (mode == EMath ?
			  QObject::tr("Enter formula (LaTeX source)") :
			  QObject::tr("Enter text (LaTeX source)")), text);

  if (dialog->exec() == QDialog::Accepted) {
    text = text.stripWhiteSpace();
    if (!text.isEmpty()) {
      IpeString data = IpeQ(text);
      if (mode == EMath)
	data = IpeString("$") + data + "$";
      IpeText *obj = new IpeText(objectAdder->ObjAdd_Attributes(),
				 data, pos, objectAdder->ObjAdd_StyleSheet());
      objectAdder->ObjAdd_AddObject(obj);
      canvas->Update();
    }
  }
}

bool CreateText::Edit(IpeText *obj, const IpeStyleSheet *sheet)
{
  bool ok = false;
  QString text;
  text = QIpe(obj->Text());
  DialogCreateText *dialog =
    new DialogCreateText(0, QObject::tr("Edit text object"),
			 QObject::tr("Edit LaTeX source"), text);
  if (dialog->exec() == QDialog::Accepted) {
    text = text.stripWhiteSpace();
    if (!text.isEmpty()) {
      obj->SetText(IpeQ(text));
      return true;
    }
  }
  return false;
}

void CreateText::Draw(QPaintEvent *ev, QPainter *qPainter) const
{
  qPainter->setPen(CreateColor());
  OverlayPainter painter(iCanvas->StyleSheet(), qPainter);
  painter.Transform(iCanvas->CanvasTfm());
  IpeVector d(0, -20.0);
  painter.BeginClosedPath(iV[0]);
  painter.LineTo(iV[1]);
  painter.LineTo(iV[1] + d);
  painter.LineTo(iV[0] + d);
  painter.EndClosedPath();
  painter.DrawPath();
}

void CreateText::MouseMove(QMouseEvent *ev)
{
  iV[1].iX = iCanvas->Pos().iX;
  iCanvas->UpdateOverlay();
}

void CreateText::MousePress(QMouseEvent *ev)
{
  IpeScalar wid = iCanvas->Pos().iX - iV[0].iX;
  if (wid < 0) {
    iV[0].iX += wid;
    wid = -wid;
  }
  QString text;
  DialogCreateText *dialog =
    new DialogCreateText(iCanvas, QObject::tr("Create text object"),
			 QObject::tr("Enter LaTeX source"), text);
  if (dialog->exec() == QDialog::Accepted) {
    text = text.stripWhiteSpace();
    if (!text.isEmpty()) {
      IpeText *obj = new IpeText(iObjectAdder->ObjAdd_Attributes(),
				 IpeQ(text), iV[0], wid,
				 iObjectAdder->ObjAdd_StyleSheet());
      iObjectAdder->ObjAdd_AddObject(obj);
    }
  }
  iCanvas->FinishOverlay();
}

// --------------------------------------------------------------------

TextWidthChanger::TextWidthChanger(Canvas *canvas, IpeText *text)
  : Overlay(canvas), iObj(text)
{
  iV[0] = iObj->Matrix() * iObj->Position();
  iV[1] = iV[0];
  iV[1].iX = iV[0].iX + iObj->Width();
  iDragging = false;
}

void TextWidthChanger::Draw(QPaintEvent *ev, QPainter *qPainter) const
{
  qPainter->setPen(Qt::magenta);
  OverlayPainter painter(iCanvas->StyleSheet(), qPainter);
  painter.Transform(iCanvas->CanvasTfm());
  IpeVector d(0, -20.0);
  painter.BeginClosedPath(iV[0]);
  painter.LineTo(iV[1]);
  painter.LineTo(iV[1] + d);
  painter.LineTo(iV[0] + d);
  painter.EndClosedPath();
  painter.DrawPath();
}

void TextWidthChanger::MouseRelease(QMouseEvent *ev)
{
  IpeScalar wid = IpeAbs(iCanvas->Pos().iX - iV[0].iX);
  iObj->SetWidth(wid);
  iCanvas->FinishOverlay();
}

void TextWidthChanger::MousePress(QMouseEvent *ev)
{
  iMouseDown = iCanvas->Pos();
  iDragging = true;
  iCanvas->UpdateOverlay();
}

void TextWidthChanger::MouseMove(QMouseEvent *ev)
{
  if (iDragging) {
    IpeScalar d = iCanvas->Pos().iX - iMouseDown.iX;
    iV[1].iX = iV[0].iX + iObj->Width() + d;
    iCanvas->UpdateOverlay();
  }
}

// --------------------------------------------------------------------
