// --------------------------------------------------------------------
// Main function
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2003  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "ipepdfdoc.h"
#include "pdflatex.h"
#include "appui.h"
#include "qipe.h"
#include "pdflib.h"
#include "debugwin.h"

#include <qapplication.h>
#include <qdir.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qtextcodec.h>

#include <cstdio>
#include <cstdlib>

#ifdef WIN32
#include <windows.h>
#endif

void StartManual(QString url)
{
  //HelpWindow *manual = HelpWindow::Static(url);
  //manual->raise();
#ifdef WIN32
  ShellExecute(0, "open", url.local8Bit(), 0, 0, 0);
#else
  QString browser = SPreferences::Static()->iBrowser;
  QString s = browser.arg(url);
  system((s + " &").local8Bit());
#endif
}

int toxml(const char *src, const char *xml)
{
  IpePdfDocument *doc =
    IpePdfDocument::New(src, IpePdfDocument::FileFormat(src));

  if (!doc) {
    fprintf(stderr, "Could not read Ipe file '%s'\n", src);
    return 1;
  }

  doc->Save(xml, IpePdfDocument::EXml);
  delete doc;
  return 0;
}

int runlatex(IpePdfDocument *doc)
{
  switch (doc->RunLatex(true)) {
  case IpePdfDocument::ErrNoText:
    fprintf(stderr, "No text objects in document, no need to run Pdflatex\n");
    return 0;
  case IpePdfDocument::ErrNoDir:
    fprintf(stderr, "Directory '%s' does not exist and cannot be created\n",
	    ((const char *) SPreferences::Static()->iLatexDir.local8Bit()));
    return 1;
  case IpePdfDocument::ErrWritingSource:
    fprintf(stderr, "Error writing Latex source.\n");
    return 1;
  case IpePdfDocument::ErrOldPdfLatex:
    fprintf(stderr, "Your installed version of Pdflatex is too old.\n");
    return 1;
  case IpePdfDocument::ErrLatex:
    fprintf(stderr, "There were errors trying to run Pdflatex\n");
    return 1;
  case IpePdfDocument::ErrNone:
  default:
    fprintf(stderr, "Pdflatex was run sucessfully.\n");
    return 0;
  }
}

int topng(const char *src, const char *dst, int pageNum, double zoom)
{
  IpePdfDocument *doc =
    IpePdfDocument::New(src, IpePdfDocument::FileFormat(src));

  if (!doc) {
    fprintf(stderr, "Could not read Ipe file '%s'\n", src);
    return 1;
  }
  if (pageNum < 1 || pageNum > doc->size()) {
    fprintf(stderr,
	    "The document contains %d pages, cannot convert page %d.\n",
	    doc->size(), pageNum);
    delete doc;
    return 1;
  }

  if (runlatex(doc)) {
    delete doc;
    return 1;
  }

  const IpePage *page = (*doc)[pageNum - 1];
  QPixmap pixmap = AppUi::ToPng(doc, page, zoom);
  delete doc;

  fprintf(stderr, "src: %s dst: %s res: %g\n", src, dst, zoom);
  if (!pixmap.save(dst, "PNG")) {
    fprintf(stderr, "Error saving bitmap '%s'\n", dst);
    return 1;
  }
  return 0;
}

int topdforps(const char *src, const char *dst, IpePdfDocument::TFormat fm)
{
  IpePdfDocument *doc =
    IpePdfDocument::New(src, IpePdfDocument::FileFormat(src));

  if (!doc) {
    fprintf(stderr, "Could not read Ipe file '%s'\n", src);
    return 1;
  }
  if (fm == IpePdfDocument::EEps && doc->size() > 1) {
    fprintf(stderr,
	    "The document contains %d pages, cannot convert to EPS format.\n",
	    doc->size());
    delete doc;
    return 1;
  }

  fprintf(stderr, "Document %s has %d pages\n", src, doc->size());

  int exitCode = runlatex(doc);
  if (!exitCode)
    doc->Save(dst, fm);
  delete doc;
  return exitCode;
}

void usage()
{
  fprintf(stderr, "Usage:\n (1) ipe [ Qt options ] [ filename ]\n");
  fprintf(stderr, " (2) ipe -toxml infile file.xml\n");
  fprintf(stderr, " (3) ipe -topdf infile file.pdf\n");
  fprintf(stderr, " (4) ipe -toeps infile file.eps\n");
  fprintf(stderr, " (5) ipe -tops  infile file.ps\n");
  fprintf(stderr, " (6) ipe -topng <page> <resolution> infile file.png\n");
  fprintf(stderr, " (7) ipe -help\n");
  fprintf(stderr, "Form (1) starts the graphical user interface.\n");
  fprintf(stderr,
	  "Forms (2-5) convert between Ipe file formats without starting\n");
  fprintf(stderr,
	  " the graphical user interface, running Pdflatex if necessary.\n");
  fprintf(stderr,"Form (6) saves a single page as a bitmap\n");
  fprintf(stderr," (<resolution> is the number of pixels per inch).\n");
  fprintf(stderr, "Form (7) prints this  message.\n");
  exit(1);
}

// The main function just starts the QApplication
int main(int argc, char *argv[])
{
  // Check Ipelib version mismatch
  if (IpelibVersion() != IPELIB_VERSION) {
    fprintf(stderr,
	    "Ipe has been compiled with header files for Ipelib %d\n"
	    "but is (probably dynamically) linked against libipe %d.\n"
	    "Check with 'ldd ipe' which libipe is being loaded, and either "
	    "replace it by the correct version or set LD_LIBRARY_PATH.",
	    IPELIB_VERSION, IpelibVersion());
    exit(99);
  }

  // initialize Xpdflib
  globalParams = new GlobalParams(0);

  if (argc >= 2) {
    QString s(argv[1]);
    if (s == "-help" || s == "--help") {
      usage();
    } else if (s == "-toxml") {
      if (argc != 4)
	usage();
      return toxml(argv[2], argv[3]);
    } else if (s == "-topdf") {
      if (argc != 4)
	usage();
      return topdforps(argv[2], argv[3], IpePdfDocument::EPdf);
    } else if (s == "-tops") {
      if (argc != 4)
	usage();
      return topdforps(argv[2], argv[3], IpePdfDocument::EPs);
    } else if (s == "-toeps") {
      if (argc != 4)
	usage();
      return topdforps(argv[2], argv[3], IpePdfDocument::EEps);
    } else if (s == "-topng") {
      if (argc < 6)
	usage();
      int pageNum = IpeLex(IpeString(argv[2])).GetInt();
      double dpi = IpeLex(IpeString(argv[3])).GetDouble();
      const char *src = argv[4];
      const char *dst = argv[5];
      // topng needs a QApplication (it creates a pixmap)
      QApplication a(argc, argv);
      return topng(src, dst, pageNum, dpi / 72.0);
    }
  }

  QApplication a( argc, argv );
  QApplication::setFont(SPreferences::Static()->iFont, true);

  QString lang = SPreferences::Static()->iLanguage;
  DBG << "Current locale is " << QTextCodec::locale() << "\n";

  QTranslator qt(0);
  QTranslator qt1(0);
  if (lang != QString::null) {
    if (qt.load("ipe-" + lang, SPreferences::Static()->iLangDir))
      a.installTranslator(&qt);
    if (qt1.load("keys-" + lang, SPreferences::Static()->iLangDir))
      a.installTranslator(&qt1);
  }

  QTranslator qt2(0);
  if (qt2.load(SPreferences::Static()->iKeysFileName)) {
    DBG << "Loaded personal key definitions\n";
    a.installTranslator(&qt2);
  }

  if (a.argc() > 2)
    usage();

  AppUi *appui = new AppUi;

  if (a.argc() != 2 || !appui->Load(a.argv()[1]))
    appui->NewDoc();

  if ( QApplication::desktop()->width() > appui->width() + 10
       && QApplication::desktop()->height() > appui->height() +30 )
    appui->show();
  else
    appui->showMaximized();
  appui->show();
  appui->FitPage();
  QObject::connect(&a, SIGNAL(lastWindowClosed()),
		   &a, SLOT(quit()));
  return a.exec();
}
