// --------------------------------------------------------------------
// Overlay factory for object creators
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2003  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "overlay.h"
#include "createpath.h"
#include "createarc.h"
#include "createtext.h"
#include "ipemark.h"
#include "canvas.h"
#include "prefs.h"

// --------------------------------------------------------------------

static const char *strings[4 * OverlayFactory::Num] = {
  QT_TR_NOOP("Select objects (with Shift: non-destructively)"),
  "s", "Mode|Select",
  "select",
  QT_TR_NOOP("Move objects (with Shift: horizontal/vertical)"),
  "m", "Mode|Move",
  "move",
  QT_TR_NOOP("Rotate objects"),
  "r", "Mode|Rotate",
  "rotate",
  QT_TR_NOOP("Stretch objects (with Shift: scale objects)"),
  "e", "Mode|Stretch",
  "stretch",
  QT_TR_NOOP("Pan the canvas"),
  "none", "Mode|Pan",
  "pan",
  QT_TR_NOOP("Text labels"),
  "t", "Mode|Text labels",
  "label",
  QT_TR_NOOP("Mathematical symbols"),
  "$", "Mode|Mathematical symbols",
  "math",
  QT_TR_NOOP("Paragraphs"),
  "g", "Mode|Paragraphs",
  "paragraph",
  QT_TR_NOOP("Marks"),
  "k", "Mode|Marks",
  "marks",
  QT_TR_NOOP("Rectangles (with Shift: squares)"),
  "b", "Mode|Rectangles",
  "rectangles",
  QT_TR_NOOP("Lines and polylines"),
  "l", "Mode|Lines",
  "lines",
  QT_TR_NOOP("Polygons"),
  "p", "Mode|Polygons",
  "polygons",
  QT_TR_NOOP("Circular arcs (by center, right and left point)"),
  "a", "Mode|Circular arcs (center, right, left)",
  "arc1",
  QT_TR_NOOP("Circular arcs (by center, left and right point)"),
  "Shift+A", "Mode|Circular arcs (center, left, right)",
  "arc2",
  QT_TR_NOOP("Circular arcs (by 3 points)"),
  "Shift+Y", "Mode|Circular arcs (3 points)",
  "arc3",
  QT_TR_NOOP("Circles (by center and radius)"),
  "c", "Mode|Circles (center, radius)",
  "circle1",
  QT_TR_NOOP("Circles (by diameter)"),
  "Shift+D", "Mode|Circles (diameter)",
  "circle2",
  QT_TR_NOOP("Circles (by 3 points)"),
  "y", "Mode|Circles (3 points)",
  "circle3",
  QT_TR_NOOP("Splines"),
  "i", "Mode|Splines",
  "splines",
  QT_TR_NOOP("Splinegons"),
  "Shift+I", "Mode|Splinegons",
  "splinegons" };

const char *whatsThisText = QT_TR_NOOP(
"<qt>With these buttons you select the current Ipe <em>mode</em>. "
"The current mode determines the function of the left mouse button. "
"<p>The leftmost five buttons set modes for selecting and transforming "
"objects.  The remaining buttons set modes for creating various kinds "
"of objects.</p></qt>");

// --------------------------------------------------------------------

/*! \class OverlayFactory
  \brief Creates the various overlays for creating objects.
*/

OverlayFactory::OverlayFactory(QMouseEvent *ev, Canvas *canvas,
			       IpePage *page, ObjectAdder *objectAdder)
  : iEvent(ev), iCanvas(canvas), iPage(page), iObjectAdder(objectAdder)
{
  // nothing
}

Overlay *OverlayFactory::StartTransforming(Transforming::TType type)
{
  Transforming *t = new Transforming(iEvent, iPage, iCanvas,
				     type, iObjectAdder);
  if (!t->FindObjects(SPreferences::Static()->iSelectDistance)) {
    delete t;
    iCanvas->Message(QObject::tr("No object selected"));
    return 0;
  }
  return t;
}

//! Create a new overlay for the canvas.
/*! The argument \a locked indicates whether the current layer is locked.
  In that case, CreateOverlay will refuse to generate object-creating
  overlays, and issue an error message. */
void OverlayFactory::CreateOverlay(int index, bool locked)
{
  iCanvas->SetOverlay(MakeOverlay(index, locked));
}

//! Return appropriate object creator.
Overlay *OverlayFactory::MakeOverlay(int index, bool locked)
{
  // check locked condition
  if (locked) {
    switch (index) {
    case ESelecting:
    case EPanning:
    case EMoving:
    case ERotating:
    case EStretching:
      break;
    default:
      iCanvas->Message(QObject::tr("Current layer is locked"));
      return 0;
    }
  }

  switch (index) {
  case ESelecting:
    return new Selecting(iEvent, iPage, iCanvas,
			 (iEvent->state() & QMouseEvent::ShiftButton),
			 SPreferences::Static()->iSelectDistance,
			 iObjectAdder);
  case EPanning:
    return new Panning(iEvent, iPage, iCanvas);
  case EMoving:
    return StartTransforming(Transforming::EMove);
  case ERotating:
    return StartTransforming(Transforming::ERotate);
  case EStretching:
    if (iEvent->state() & QMouseEvent::ShiftButton)
      return StartTransforming(Transforming::EScale);
    else
      return StartTransforming(Transforming::EStretch);
  case ERectangle:
    return new CreateRectangle(iEvent, iCanvas, iObjectAdder);
  case EPolyline:
    return new CreatePath(iEvent, iCanvas, iObjectAdder, false, false);
  case EPolygon:
    return new CreatePath(iEvent, iCanvas, iObjectAdder, true, false);
  case ESpline:
    return new CreatePath(iEvent, iCanvas, iObjectAdder, false, true);
  case ESplinegon:
    return new CreateSplinegon(iEvent, iCanvas, iObjectAdder);
  case ECircle1:
    return new CreateCircle(iEvent, iCanvas, iObjectAdder,
			    CreateCircle::ECenterRadius);
  case ECircle2:
    return new CreateCircle(iEvent, iCanvas, iObjectAdder,
			    CreateCircle::EDiameter);
  case ECircle3:
    return new CreateCircle(iEvent, iCanvas, iObjectAdder,
			    CreateCircle::E3Points);
  case EArc1:
    return new CreateArc(iEvent, iCanvas, iObjectAdder,
			 CreateArc::ECenterRightLeft);
  case EArc2:
    return new CreateArc(iEvent, iCanvas, iObjectAdder,
			 CreateArc::ECenterLeftRight);
  case EArc3:
    return new CreateArc(iEvent, iCanvas, iObjectAdder,
			 CreateArc::E3Points);
  case ELabel:
    CreateText::New(iEvent, iCanvas, iObjectAdder, CreateText::ELabel);
    return 0;
  case EMath:
    CreateText::New(iEvent, iCanvas, iObjectAdder, CreateText::EMath);
    return 0;
  case EParagraph:
    return new CreateText(iEvent, iCanvas, iObjectAdder);
  case EMark:
    {
      iObjectAdder->ObjAdd_AddObject
	(new IpeMark(iObjectAdder->ObjAdd_Attributes(), iCanvas->Pos()));
      iCanvas->Update();
      return 0;
    }
  default:
    return 0;
  }
}

const char *OverlayFactory::PixmapName(int index)
{
  return strings[4 * index + 3];
}

const char *OverlayFactory::Tooltip(int index)
{
  return strings[4 * index + 0];
}

const char *OverlayFactory::Shortcut(int index)
{
  return strings[4 * index + 1];
}

const char *OverlayFactory::Context(int index)
{
  return strings[4 * index + 2];
}

const char *OverlayFactory::WhatsThis(int index)
{
  return whatsThisText;
}

// --------------------------------------------------------------------
