// -*- C++ -*-
// --------------------------------------------------------------------
// Overlay for the canvas --- shows moving objects etc.
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2003  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef OVERLAY_H
#define OVERLAY_H

#include "ipegeo.h"
#include "ipepainter.h"
#include "ipepage.h"

#include <qpoint.h>
#include <qcolor.h>

// --------------------------------------------------------------------

class IpePage;
class IpeStyleSheet;
class QPainter;
class QMouseEvent;
class QPaintEvent;
class QKeyEvent;
class Canvas;
class IpeAllAttributes;

class OverlayPainter : public IpePainter {
public:
  OverlayPainter(const IpeStyleSheet *sheet, QPainter *painter);
  virtual ~OverlayPainter() { }

  virtual void BeginPath(const IpeVector &v);
  virtual void BeginClosedPath(const IpeVector &v);
  virtual void LineTo(const IpeVector &v);
  virtual void CurveTo(const IpeVector &v1, const IpeVector &v2,
		       const IpeVector &v3);
  virtual void EndPath();
  virtual void EndClosedPath();

  virtual void DrawBitmap(IpeBitmap bitmap);
  virtual void DrawText(const IpeText *text);

private:
  QPainter *iPainter;
  QPoint iPathBegin;
  QPoint iPos;
};

// --------------------------------------------------------------------

class Overlay {
public:
  virtual ~Overlay();

  //! Redraw the current overlay.
  virtual void Draw(QPaintEvent *ev, QPainter *painter) const = 0;

  virtual void MousePress(QMouseEvent *ev);
  virtual void MouseRelease(QMouseEvent *ev);
  virtual void MouseMove(QMouseEvent *ev);
  virtual void KeyPress(QKeyEvent *ev);

  inline QColor CreateColor() const { return QColor("navy"); }

 protected:
  Overlay(Canvas *canvas);

 protected:
  Canvas *iCanvas;
};

// --------------------------------------------------------------------

/*! \class ObjectAdder
  \brief Interface for adding objects to the current document.

  Provides services to (object-creating) overlays.
*/

class ObjectAdder {
public:
  virtual void ObjAdd_AddObject(IpeObject *obj) = 0;
  virtual const IpeAllAttributes &ObjAdd_Attributes() = 0;
  virtual const IpeStyleSheet *ObjAdd_StyleSheet() = 0;
  virtual void ObjAdd_AddUndoItem(IpePage *page, QString s) = 0;
  virtual void ObjAdd_AddUndoItem(IpePage::iterator it, IpePage *page,
				  const IpePgObject &original, QString s) = 0;
};

// --------------------------------------------------------------------

class Panning : public Overlay {
 public:
  Panning(QMouseEvent *ev, const IpePage *page, Canvas *canvas);

  virtual void Draw(QPaintEvent *ev, QPainter *painter) const;
  virtual void MouseRelease(QMouseEvent *ev);
  virtual void MouseMove(QMouseEvent *ev);

 private:
  const IpePage *iPage;
  IpeVector iMouseDown;
  IpeVector iPan;
};

// --------------------------------------------------------------------

class Selecting : public Overlay {
 public:
  Selecting(QMouseEvent *ev, IpePage *page, Canvas *canvas,
	    bool nonDestructive, int selectDistance,
	    ObjectAdder *objAdd);

  virtual void Draw(QPaintEvent *ev, QPainter *painter) const;
  virtual void MouseRelease(QMouseEvent *ev);
  virtual void MouseMove(QMouseEvent *ev);
  virtual void KeyPress(QKeyEvent *ev);

private:
  void EnsurePrimary();

private:
  class SelectCompare;

  IpePage *iPage;
  ObjectAdder *iObjectAdder;
  bool iNonDestructive;
  int iSelectDistance;
  IpeVector iMouseDown;
  struct SObj {
    IpePage::iterator iIt;
    double iDistance;
  };
  std::vector<SObj> iObjs;
  int iCur;

  bool iDragging;
  IpeVector iCorner;
};

// --------------------------------------------------------------------

class Transforming : public Overlay {
 public:
  enum TType { EMove, EScale, EStretch, ERotate };

  Transforming(QMouseEvent *ev, IpePage *page, Canvas *canvas, TType type,
	       ObjectAdder *objectAdder);
  bool FindObjects(int selectDistance);

  virtual void Draw(QPaintEvent *ev, QPainter *painter) const;
  virtual void MouseMove(QMouseEvent *ev);
  virtual void MouseRelease(QMouseEvent *ev);

 private:
  void Compute(const IpeVector &v);
  bool UpdateCloseSelection(int selectDistance);

 private:
  IpePage *iPage;
  ObjectAdder *iObjectAdder;
  TType iType;
  bool iWithShift;
  IpeVector iMouseDown;
  IpeMatrix iTransform;
  IpeVector iOrigin;
};

// --------------------------------------------------------------------

class OverlayFactory {
public:
  enum { ESelecting, EMoving, ERotating, EStretching, EPanning,
	 ELabel, EMath, EParagraph, EMark,
	 ERectangle, EPolyline, EPolygon,
	 EArc1, EArc2, EArc3,
	 ECircle1, ECircle2, ECircle3,
	 ESpline, ESplinegon, Num };

  OverlayFactory(QMouseEvent *ev, Canvas *canvas,
		 IpePage *page, ObjectAdder *objectAdder);
  void CreateOverlay(int index, bool locked = false);
  Overlay *MakeOverlay(int index, bool locked);
  static const char *PixmapName(int index);
  static const char *Tooltip(int index);
  static const char *WhatsThis(int index);
  static const char *Shortcut(int index);
  static const char *Context(int index);
private:
  Overlay *StartTransforming(Transforming::TType type);
private:
  QMouseEvent *iEvent;
  Canvas *iCanvas;
  IpePage *iPage;
  ObjectAdder *iObjectAdder;
};

// --------------------------------------------------------------------
#endif
