// --------------------------------------------------------------------
// Interface for Qt and Ipelib
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2003  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "qipe.h"
#include "ipestyle.h"

#include <qfile.h>
#include <qpainter.h>

// --------------------------------------------------------------------

/*! \class IpeQStream
  \brief Stream writing into an open QIODevice.
*/

//! Constructor.
IpeQStream::IpeQStream(QIODevice *dev)
  : iDevice(dev)
{
  // nothing
}

void IpeQStream::PutChar(char ch)
{
  iDevice->putch(ch);
}

void IpeQStream::PutString(IpeString s)
{
  for (int i = 0; i < s.size(); ++i)
    iDevice->putch(s[i]);
}

void IpeQStream::PutCString(const char *s)
{
  while (*s) iDevice->putch(*s++);
}

void IpeQStream::PutRaw(const char *data, int size)
{
  while (size-- > 0)
    iDevice->putch(*data++);
}

// --------------------------------------------------------------------

/*! \class XmlQStringSource
  \brief IpeXmlDataSource reading from a QString.
*/

//! Construct XML source.
XmlQStringSource::XmlQStringSource(QString str)
{
  iData = IpeQ(str);
  iPos = 0;
}

//! Get one character.
int XmlQStringSource::GetChar()
{
  if (iPos < iData.size())
    return iData[iPos++];
  else
    return EOF;
}

// --------------------------------------------------------------------

/*! \class XmlQSource
  \brief Data source for parsing XML from a QIODevice.
*/

XmlQSource::XmlQSource(QIODevice *dev, bool skipPercent)
  : iSkipPercent(skipPercent), iDev(dev)
{
  // nothing
}

int XmlQSource::GetChar()
{
  int ch = iDev->getch();
  if (iSkipPercent && ch == '\n')
    iDev->getch(); // skip '%'
  return (ch < 0) ? EOF : ch;
}

// --------------------------------------------------------------------

QPixmap ColorPixmap(IpeAttribute sym, const IpeStyleSheet *sheet)
{
  QPixmap pixmap;
  pixmap.resize(14, 14);
  if (sym.IsVoid()) {
    pixmap.fill(Qt::white);
    QPainter p(&pixmap);
    p.drawRect(0, 0, 14, 14);
    p.drawLine(0, 0, 14, 14);
    p.drawLine(0, 14,14,  0);
  } else {
    IpeColor col = sheet->Repository()->ToColor(sheet->Find(sym));
    QColor color(int(col.iRed * 255),
		 int(col.iGreen * 255),
		 int(col.iBlue * 255));
    pixmap.fill(color);
  }
  return pixmap;
}

// --------------------------------------------------------------------
