// --------------------------------------------------------------------
// IpeImlParser
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2003  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "ipeiml.h"
#include "ipeobj.h"
#include "ipegroup.h"
#include "ipepage.h"
#include "ipefactory.h"
#include "ipestyle.h"
// #include "ipeimage.h"

// --------------------------------------------------------------------

/*! \class IpeImlParser
  \ingroup high
  \brief Ipe XML Parser.

  A recursive descent parser for the Ipe XML streams.

  After experimenting with various XML parsing frameworks, this turned
  out to work best for Ipe.

*/

IpeImlParser::IpeImlParser(IpeXmlDataSource &source,
			   IpeRepository *rep)
  : IpeXmlParser(source), iRepository(rep)
{
  // nothing
}

//! XML contents can refer to data in PDF.
/*! If the XML stream is embedded in a PDF file, XML contents can
  refer to PDF objects.  A derived parser must implement this method
  to access PDF data.

  It is assumed that PDF object \a objNum is a stream.  Its contents
  is returned as a single (possibly very long) string.
*/
IpeString IpeImlParser::PdfStream(int objNum)
{
  return IpeString();
}

//! Read a complete Ipe document from IML stream.
/*! Sets \a requires to the Ipelib version required by the document,
  e.g. 60100 for the Ipelib of Ipe 6.1.
  Returns false if that is larger than the version of this Ipelib. */
bool IpeImlParser::ParseDocument(IpeDocument &doc, int &requires)
{
  requires = 0;
  IpeDocument::SProperties properties = doc.Properties();

  IpeString tag = ParseToTag();
  if (tag == "?xml") {
    IpeXmlAttributes attr;
    if (!ParseAttributes(attr))
      return false;
    tag = ParseToTag();
  }
  if (tag != "ipe")
    return false;

  IpeXmlAttributes attr;
  if (!ParseAttributes(attr))
    return false;

  IpeLex versLex(attr["version"]);
  versLex >> requires;
  if (requires > IPELIB_VERSION)
    return false;

  attr.Has("creator", properties.iCreator);

  tag = ParseToTag();
  if (tag == "info") {
    IpeXmlAttributes att;
    if (!ParseAttributes(att))
      return false;

    properties.iTitle = att["title"];
    properties.iAuthor = att["author"];
    properties.iSubject = att["subject"];
    properties.iKeywords = att["keywords"];
    properties.iFullScreen = (att["pagemode"] == "fullscreen");
    properties.iCropBox = (att["bbox"] == "cropbox");
    properties.iCreated = att["created"];
    properties.iModified = att["modified"];

    tag = ParseToTag();
  }
  if (tag == "preamble") {
    IpeXmlAttributes att;
    if (!ParseAttributes(att))
      return false;
    if (!ParsePCDATA("preamble", properties.iPreamble))
      return false;
    tag = ParseToTag();
  }
  doc.SetProperties(properties);

  // bottom style sheet in cascade is always "standard"
  IpeStyleSheet *sheet = IpeStyleSheet::Standard(iRepository);
  if (!sheet)
    return false;
  doc.SetStyleSheet(sheet);

  while (tag == "ipestyle") {
    IpeStyleSheet *sheet = new IpeStyleSheet(iRepository);
    if (!ParseStyle(*sheet)) {
      delete sheet;
      return false;
    }
    sheet->SetCascade(doc.GetStyleSheet());
    doc.SetStyleSheet(sheet);
    tag = ParseToTag();
  }

  while (tag == "bitmap") {
    if (!ParseBitmap())
      return false;
    tag = ParseToTag();
  }

  while (tag == "page") {
    // read one page
    IpePage *page = new IpePage;
    doc.push_back(page);
    if (!ParsePage(*page))
      return false;
    tag = ParseToTag();
  }
  IpeAttributeSeq seq;
  if (!doc.NotifyStyle(seq))
    return false;
  doc.SetEdited(false);
  return (tag == "/ipe");
}

//! Parse an IpeBitmap.
/*! On calling, stream must be just past \c bitmap. */
bool IpeImlParser::ParseBitmap()
{
  IpeXmlAttributes att;
  if (!ParseAttributes(att))
    return false;
  IpeString objNumStr;
  if (att.Slash() && att.Has("pdfObject", objNumStr)) {
    int objNum = IpeLex(objNumStr).GetInt();
    IpeBitmap bitmap(att, PdfStream(objNum), true);
    iBitmaps.push_back(bitmap);
  } else {
    IpeString bits;
    if (!ParsePCDATA("bitmap", bits))
      return false;
    IpeBitmap bitmap(att, bits, false);
    iBitmaps.push_back(bitmap);
  }
  return true;
}

//! Parse an IpePage.
/*! On calling, stream must be just past \c page. */
bool IpeImlParser::ParsePage(IpePage &page)
{
  IpeXmlAttributes att;
  if (!ParseAttributes(att))
    return false;
  IpeString mediabox;
  if (!att.Has("mediabox", mediabox))
    return false;
  IpeLex lex(mediabox);
  IpeVector mbmin;
  IpeVector mbmax;
  lex >> mbmin.iX >> mbmin.iY >> mbmax.iX >> mbmax.iY;
  page.SetMediaBox(IpeRect(mbmin, mbmax));
  IpeString tag = ParseToTag();
  while (tag == "layer") {
    IpeXmlAttributes att;
    if (!ParseAttributes(att))
      return false;
    IpeLayer layer(att);
    page.AddLayer(layer);
    tag = ParseToTag();
  }
  // default layer: 'alpha'
  if (page.CountLayers() == 0) {
    IpeLayer layer("alpha");
    page.AddLayer(layer);
  }
  while (tag == "view") {
    IpeXmlAttributes att;
    if (!ParseAttributes(att))
      return false;
    IpePresentation pres(att);
    page.AddPresentation(pres);
    tag = ParseToTag();
  }
  int currentLayer = 0;
  IpeString layerName;
  for (;;) {
    if (tag == "/page") {
      return true;
    }
    IpeObject *obj = ParseObject(tag, &page, &currentLayer);
    if (!obj)
      return false;
    page.push_back(IpePgObject(IpePgObject::ENone, currentLayer, obj));
    tag = ParseToTag();
  }
}

//! Parse an \c <ipepage> element (used on clipboard).
IpePage *IpeImlParser::ParsePageSelection()
{
  IpeString tag = ParseToTag();
  if (tag != "ipepage")
    return false;
  IpeXmlAttributes attr;
  if (!ParseAttributes(attr))
    return 0;
  tag = ParseToTag();

  while (tag == "bitmap") {
    if (!ParseBitmap())
      return false;
    tag = ParseToTag();
  }

  if (tag != "page")
    return 0;

  IpeAutoPtr<IpePage> page(new IpePage);
  if (!ParsePage(*page))
    return 0;

  tag = ParseToTag();
  if (tag != "/ipepage")
    return 0;
  return page.Take();
}

//! Parse an \c <ipeselection> element.
/*! An IpePgObjectSeq is used to own the objects, but selection mode
  and layer are not set. */
bool IpeImlParser::ParseSelection(IpePgObjectSeq &seq)
{
  IpeString tag = ParseToTag();
  if (tag != "ipeselection")
    return false;
  IpeXmlAttributes attr;
  if (!ParseAttributes(attr))
    return 0;
  tag = ParseToTag();

  while (tag == "bitmap") {
    if (!ParseBitmap())
      return false;
    tag = ParseToTag();
  }

  for (;;) {
    if (tag == "/ipeselection")
      return true;
    IpeObject *obj = ParseObject(tag);
    if (!obj)
      return false;
    seq.push_back(IpePgObject(IpePgObject::ENone, 0, obj));
    tag = ParseToTag();
  }
}

//! Parse an IpeObject.
/*! On calling, stream must be just past the tag. */
IpeObject *IpeImlParser::ParseObject(IpeString tag, IpePage *page,
				     int *currentLayer)
{
  if (tag[0] == '/')
    return 0;

  IpeXmlAttributes attr;
  if (!ParseAttributes(attr))
    return 0;

  IpeString layer;
  if (page && currentLayer && attr.Has("layer", layer)) {
    for (int i = 0; i < page->CountLayers(); ++i) {
      if (page->Layer(i).iName == layer) {
	*currentLayer = i;
	break;
      }
    }
  }

  if (tag == "group") {
    IpeGroup group(iRepository, attr);
    for (;;) {
      IpeString tag = ParseToTag();
      if (tag == "/group") {
	return new IpeGroup(group);
      }
      IpeObject *obj = ParseObject(tag);
      if (!obj)
	return 0;
      group.push_back(obj);
    }
  }

  IpeString pcdata;
  if (!attr.Slash() && !ParsePCDATA(tag, pcdata))
    return 0;
  IpeString bitmapId;
  if (tag == "image" && attr.Has("bitmap", bitmapId)) {
    int objNum = IpeLex(bitmapId).GetInt();
    IpeBitmap bitmap;
    for (std::vector<IpeBitmap>::const_iterator it = iBitmaps.begin();
	 it != iBitmaps.end(); ++it) {
      if (it->ObjNum() == objNum) {
	bitmap = *it;
	break;
      }
    }
    assert(!bitmap.IsNull());
    return IpeObjectFactory::CreateImage(iRepository, tag, attr, bitmap);
  } else
    return IpeObjectFactory::CreateObject(iRepository, tag, attr, pcdata);
}

//! Parse a style sheet.
/*! On calling, stream must be just past the style tag. */
bool IpeImlParser::ParseStyle(IpeStyleSheet &sheet)
{
  IpeXmlAttributes att;
  if (!ParseAttributes(att))
    return false;
  IpeString name;
  if (att.Has("name", name))
    sheet.SetName(name);

  IpeString tag = ParseToTag();
  while (tag != "/ipestyle") {
    if (tag == "template") {
      if (!ParseAttributes(att))
	return false;
      IpeString tag1 = ParseToTag();
      IpeObject *obj = ParseObject(tag1);
      if (!obj)
	return false;
      IpeString name = att["name"];
      if (name.empty())
	return false;
      IpeAttribute attr =
	iRepository->MakeSymbol(IpeAttribute::ETemplate, name);
      sheet.AddTemplate(attr, obj);
      if (ParseToTag() != "/template")
	return false;
    } else if (tag == "color") {
      if (!ParseAttributes(att) || !att.Slash())
	return false;
      IpeAttribute col = iRepository->MakeColor(att["value"]);
      if (!col.IsAbsolute())
	return false;
      IpeString name = att["name"];
      if (name.empty() || !(('a' <= name[0] && name[0] <= 'z') ||
			    ('A' <= name[0] && name[0] <= 'Z')))
	return false;
      IpeAttribute attr = iRepository->MakeSymbol(IpeAttribute::EColor, name);
      sheet.Add(attr, col);
    } else if (tag == "dashstyle") {
      if (!ParseAttributes(att) || !att.Slash())
	return false;
      IpeAttribute dash = iRepository->MakeDashStyle(att["value"]);
      if (!dash.IsAbsolute())
	return false;
      IpeString name = att["name"];
      if (name.empty() || !(('a' <= name[0] && name[0] <= 'z') ||
			    ('A' <= name[0] && name[0] <= 'Z')))
	return false;
      IpeAttribute attr =
	iRepository->MakeSymbol(IpeAttribute::EDashStyle, name);
      sheet.Add(attr, dash);
    } else if (tag == "textsize") {
      if (!ParseAttributes(att) || !att.Slash())
	return false;
      IpeString name = att["name"];
      if (name.empty())
	return false;
      IpeAttribute attr =
	iRepository->MakeSymbol(IpeAttribute::ETextSize, name);
      IpeAttribute value =
	iRepository->MakeTextSize(att["value"]);
      sheet.Add(attr, value);
    } else if (tag == "textmatrix") {
      if (!ParseAttributes(att) || !att.Slash())
	return false;
      IpeString name = att["name"];
      if (name.empty())
	return false;
      IpeAttribute attr =
	iRepository->MakeSymbol(IpeAttribute::ETextMatrix, name);
      IpeAttribute value =
	iRepository->MakeMatrix(IpeAttribute::ETextMatrix, att["value"]);
      sheet.Add(attr, value);
    } else if (tag == "media") {
      if (!ParseAttributes(att) || !att.Slash())
	return false;
      IpeAttribute media =
	iRepository->MakeVector(IpeAttribute::EMedia, att["value"]);
      IpeString name = att["name"];
      if (name.empty())
	return false;
      IpeAttribute attr = iRepository->MakeSymbol(IpeAttribute::EMedia, name);
      sheet.Add(attr, media);
    } else {
      IpeKind kind;
      if (tag == "linewidth")
	kind = IpeAttribute::ELineWidth;
      else if (tag == "marksize")
	kind = IpeAttribute::EMarkSize;
      else if (tag == "arrowsize")
	kind = IpeAttribute::EArrowSize;
      else if (tag == "grid")
	kind = IpeAttribute::EGridSize;
      else if (tag == "angle")
	kind = IpeAttribute::EAngleSize;
      else
	return false; // error
      if (!ParseAttributes(att) || !att.Slash())
	return false;
      IpeAttribute value = iRepository->MakeScalar(kind, att["value"]);
      IpeString name = att["name"];
      if (name.empty() || !value.IsAbsolute())
	return false;
      IpeAttribute attr = iRepository->MakeSymbol(kind, name);
      sheet.Add(attr, value);
    }
    tag = ParseToTag();
  }
  return true;
}

//! Parse a complete style sheet.
/*! On calling, stream must be before the 'ipestyle' tag.
  A <?xml> tag is allowed.
 */
IpeStyleSheet *IpeImlParser::ParseStyleSheet()
{
  IpeString tag = ParseToTag();
  if (tag == "?xml") {
    IpeXmlAttributes attr;
    if (!ParseAttributes(attr))
      return 0;
    tag = ParseToTag();
  }
  if (tag != "ipestyle")
    return 0;
  IpeStyleSheet *sheet = new IpeStyleSheet(iRepository);
  if (ParseStyle(*sheet))
    return sheet;
  delete sheet;
  return 0;
}

// --------------------------------------------------------------------
