// --------------------------------------------------------------------
// The reference object.
// --------------------------------------------------------------------
/*

    This file is part of the extensible drawing editor Ipe.
    Copyright (C) 1993-2003  Otfried Cheong

    Ipe is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Ipe is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License
    along with Ipe; if not, you can find it at
    "http://www.gnu.org/copyleft/gpl.html", or write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "iperef.h"
#include "ipevisitor.h"
#include "ipestyle.h"
#include "ipepainter.h"

/*! \class IpeReference
  \ingroup obj
  \brief The reference object.

  A reference reuses an object defined in an Ipe style sheet. The
  object is defined as a named template in the style sheet, and can be
  reused arbitrarily often in the document.  This is typically used
  for backgrounds on multi-page documents.

*/

//! Create from XML stream.
IpeReference::IpeReference(IpeRepository *rep, const IpeXmlAttributes &attr,
			   IpeString /* data */)
  : IpeObject(rep, attr)
{
  SetStroke(IpeAttribute()); // null stroke
  iName = rep->MakeSymbol(IpeAttribute::ETemplate, attr["name"]);
  iObject = 0;
}

//! Clone object
IpeObject *IpeReference::Clone() const
{
  return new IpeReference(*this);
}

//! Copy constructor.
IpeReference::IpeReference(const IpeReference &rhs)
  : IpeObject(rhs)
{
  iName = rhs.iName;
  iObject = rhs.iObject;
}

//! Return pointer to this object.
IpeReference *IpeReference::AsReference()
{
  return this;
}

void IpeReference::Accept(IpeVisitor &visitor) const
{
  visitor.VisitReference(this);
}

//! Save in XML format.
void IpeReference::SaveAsXml(IpePainter &painter, IpeStream &stream,
			     IpeString layer) const
{
  stream << "<ref";
  SaveAttributesAsXml(painter, stream, layer);
  stream << " name=\"" << painter.Repository()->String(Name()) << "\"/>\n";
}

//! Draw reference.
void IpeReference::Draw(IpePainter &painter) const
{
  assert(iObject);
  painter.Push();
  painter.Transform(Matrix());
  iObject->Draw(painter);
  painter.Pop();
}

void IpeReference::AddToBBox(IpeRect &box, const IpeMatrix &m) const
{
  assert(iObject);
  iObject->AddToBBox(box, m * Matrix());
}

void IpeReference::NotifyStyle(const IpeStyleSheet *sheet,
			       IpeAttributeSeq &seq) const
{
  iObject = sheet->FindTemplate(iName);
  if (!iObject && std::find(seq.begin(), seq.end(), iName) == seq.end())
    seq.push_back(iName);
}

double IpeReference::Distance(const IpeVector &v, const IpeMatrix &m,
			      double bound) const
{
  assert(iObject);
  return iObject->Distance(v, m * Matrix(), bound);
}

void IpeReference::SnapVtx(const IpeVector &mouse, const IpeMatrix &m,
			   IpeVector &pos, double &bound) const
{
  assert(iObject);
  iObject->SnapVtx(mouse, m * Matrix(), pos, bound);
}

void IpeReference::SnapBnd(const IpeVector &mouse, const IpeMatrix &m,
			   IpeVector &pos, double &bound) const
{
  assert(iObject);
  iObject->SnapBnd(mouse, m * Matrix(), pos, bound);
}

// --------------------------------------------------------------------
